$ifndef _MCP23S17_
$define _MCP23S17_
'
' Proton macros to interface with a Microchip MCP23S17 16-bit SPI I/O peripheral using the MSSP peripheral
'
' Written for the Crownhill Proton compiler version 3.5.3.1 onwards using an 18F device
'
' Author Les Johnson 23-02-2012
'
' You are free to use this code as you see fit, however, a mention of the author and compiler would be appreciated.
' The code is offered as-is and Crownhill nor the author take any responsibility for its use.
'
' MSSP SPI master configuration values
' Note. These values are for a PIC18F25K20. 
' Alter according to the device used
'
$define cSPI_SMPEND    %10000000            ' Input data sample at end of data out
$define cSPI_SMPMID    %00000000            ' Input data sample at middle of data out

$define cSPI_MODE_00   %00000000            ' Setting for SPI bus Mode 0,0
$define cSPI_MODE_01   %00000001            ' Setting for SPI bus Mode 0,1
$define cSPI_MODE_10   %00000010            ' Setting for SPI bus Mode 1,0
$define cSPI_MODE_11   %00000011            ' Setting for SPI bus Mode 1,1

$define cSPI_SSPENB    %00100000            ' Enable serial port and configures SCK, SDO, SDI

$define cSPI_FOSC_4    %00000000            ' SPI Master mode, clock = Fosc/4
$define cSPI_FOSC_16   %00000001            ' SPI Master mode, clock = Fosc/16
$define cSPI_FOSC_64   %00000010            ' SPI Master mode, clock = Fosc/64
$define cSPI_FOSC_TMR2 %00000011            ' SPI Master mode, clock = TMR2 output/2
'
' SPI interface lines
'
$ifndef MCP_CS
    $define MCP_CS PORTC.2                  ' Default location of the MCP23S17 CS line
$endif

    Symbol cMicrochipAddress = $00          ' Default address of the MCP23S17 device
'
' MCP23S17 registers
'
    Symbol cIODIRA   = $00
    Symbol cIODIRB   = $01
    Symbol cIPOLA    = $02
    Symbol cIPOLB    = $03
    Symbol cGPINTENA = $04
    Symbol cGPINTENB = $05
    Symbol cDEFVALA  = $06
    Symbol cDEFVALB  = $07
    Symbol cINTCONA  = $08
    Symbol cINTCONB  = $09
    Symbol cIOCONA   = $0A
    Symbol cIOCONB   = $0B
    Symbol cGPPUA    = $0C
    Symbol cGPPUB    = $0D
    Symbol cINTFA    = $0E
    Symbol cINTFB    = $0F
    Symbol cINTCAPA  = $10
    Symbol cINTCAPB  = $11
    Symbol cGPIOA    = $12
    Symbol cGPIOB    = $13
    Symbol cOLATA    = $14
    Symbol cOLATB    = $15
'
' Control byte for SPI operations [0100][A2][A1][A0][R=1/W=0]
'
    Symbol cControlByteMaskRead = %01000001 | (cMicrochipAddress << 1)
    Symbol cControlByteMaskWrite = %01000000 | (cMicrochipAddress << 1)

    Symbol cControl_IOCON = %00011000        ' Init value for IOCON - bits details in following comment
(*
 Register IOCON = [BANK][MIRROR][SEQOP][DISSLW][HAEN][ODR][INTPOL][-Unused]

 Bit-7 BANK:    Controls how the registers are addressed
                    1 = The registers associated With each Port are separated into different banks
                    0 = The registers are in the same bank (addresses are sequential)
 Bit-6 MIRROR:  INT Pins Mirror Bit
                    1 = The INT pins are internally connected
                    0 = The INT pins are Not connected. INTA is associated With PORTA And INTB is associated With PORTB
 Bit-5 SEQOP:   Sequential Operation mode Bit.
                    1 = Sequential operation disabled, address pointer does Not increment.
                    0 = Sequential operation Enabled, address pointer increments.
 Bit-4 DISSLW:  Slew Rate control Bit For SDA Output.
                    1 = Slew rate disabled.
                    0 = Slew rate Enabled.
 Bit-3 HAEN:    Hardware Address Enable Bit (MCP23S17 only).
                    Address pins are always Enabled on MCP23017.
                    1 = Enables the MCP23S17 address pins.
                    0 = Disables the MCP23S17 address pins.
 Bit-2 ODR:     This Bit configures the INT pin As an open-drain Output.
                    1 = Open-drain Output (overrides the INTPOL Bit).
                    0 = Active driver Output (INTPOL Bit sets the polarity).
 Bit-1 INTPOL:  This Bit sets the polarity of the INT Output pin.
                    1 = Active-High.
                    0 = Active-Low.
 Bit-0 Unimplemented: Read As 0.
*)
'
' Variables used to cache the state of Direction and Port registers
'
    Dim bShadowDirRegA As Byte
    Dim bShadowDirRegB As Byte
    Dim bShadowDataRegA As Byte
    Dim bShadowDataRegB As Byte
'
' Variables used for passing parameters to the macros (better left as System types in bankless RAM)
'
    Dim bRegAddress As Byte System
    Dim bDataByte As Byte System
    Dim bMCPtemp As bDataByte

'----------------------------------------------------------------------------------------------------
    GoTo MCP23S17_Main                                  ' Jump over the subroutines
'----------------------------------------------------------------------------------------------------
' Macro Name    : mSPI_Open
' Return Value  : None
' Parameters    : SSP peripheral setup values
' Notes         : Sets up the SSP module for master operation
'
' Syntax mSPI_Open pSyncMode , pBusMode , pSmpPhase
'
$define mSPI_Open(pSyncMode,pBusMode,pSmpPhase) _SPI_Open pSyncMode, pBusMode, pSmpPhase

_SPI_Open Macro- pSyncMode, pBusMode, pSmpPhase

    Num_SFR pSyncMode,SSPCON1           ' Select serial mode   
    Num_SFR (63 | pSmpPhase),SSPSTAT    ' Select data input sample phase

    #if(pBusMode == 0)                  ' SPI bus mode 0,0
        #if((pSyncMode & 3) == 3)       ' SPI Master mode, clock = TMR2 output/2?
            SSPSTATbits_CKE = 0         ' Data transmitted on falling edge
            SSPCON1bits_CKP = 0         ' Clock idle state low
        #else
            SSPSTATbits_CKE = 1         ' Data transmitted on rising edge
        #endif
    #endif
    #if(pBusMode == 2)                  ' SPI bus mode 1,0?
        SSPSTATbits_CKE = 1             ' Data transmitted on falling edge
        SSPCON1bits_CKP = 1             ' Clock idle state high
    #endif
    #if(pBusMode == 3)                  ' SPI bus mode 1,1?
        SSPCON1bits_CKP = 1             ' Clock idle state high
    #endif

    TRISC.4 = 1                         ' Define SDI pin as input
    TRISC.5 = 0                         ' Define SDO pin as output
    TRISC.3 = 0                         ' Define clock pin as output

    SSPCON1 = SSPCON1 | cSPI_SSPENB     ' Enable the synchronous serial port
    Endm
'----------------------------------------------------------------------------------------------------
' Write a single Byte (8-bits) To the SPI bus.
' Input     : WREG holds the byte to write to the bus
' Output    : Input data from device can be picked up by reading SSPBUF straight after the mSPI_Write macro
' Notes     : None
'
$define mSPI_Write(pDataOut) '
    WREG = pDataOut '
    _WriteSPI

_WriteSPI Macro-
    GoSub __WriteSPI
Endm

#ifMacro- _WriteSPI
__WriteSPI:
    SSPSTATBits_BF = 0              ' Empty the SSPBUF buffer flags
    PIR1bits_SSPIF = 0              ' Clear the interrupt flag
    SSPBUF = WREG                   ' Place the data to send into SSPBUF
    Btfss PIR1bits_SSPIF            ' \
    Bra $ - 2                       ' / Wait until cycle complete
    WREG = SSPBUF                   ' Read SSPBUF into WREG
    Return
#endIfMacro-
'----------------------------------------------------------------------------------------------------
' Read a single Byte from the SPI bus.
' Input     : None
' Output    : WREG holds the byte read from the device
' Notes     : None
'
$define mSPI_Read() _ReadSPI

_ReadSPI Macro- \Byte
    GoSub __ReadSPI
    Return_Wreg
Endm

#ifMacro- _ReadSPI
__ReadSPI:
    WREG = SSPBUF                   ' Empty the SSPBUF buffer and clear flags
    PIR1bits_SSPIF = 0              ' Clear interrupt flag
    SSPBUF = $FF                    ' Initiate bus cycle
    Btfss PIR1bits_SSPIF            ' \
    Bra $ - 2                       ' / Wait until cycle complete
    WREG = SSPBUF
    Return
#endIfMacro-
'----------------------------------------------------------------------------------------------------
' Write to one of the MCP23S17 registers
' Input     : bRegAddress holds the address to write
'           : bDataByte holds the data to write to the register
' Output    : None
' Notes     : None
'
$define MCP23S17_Write(pRegAddress, pDataByte) '
    bDataByte = pDataByte                      '
    WREG = pRegAddress : _Write_MCP

_Write_MCP Macro-
    GoSub __Write_MCP
Endm

#ifMacro- _Write_MCP, _WritePinPortA, _WritePinPortB, _WritePinTrisA, _WritePinTrisB
__Write_MCP:
    bRegAddress = WREG
    Clear MCP_CS
    DelayCS 1
    mSPI_Write(cControlByteMaskWrite)
    mSPI_Write(bRegAddress)
    mSPI_Write(bDataByte)
    Set MCP_CS
    Return
#endIfMacro-

'----------------------------------------------------------------------------------------------------
' Read from one of the MCP23S17 registers
' Input     : bRegAddress holds the adddress to write
' Output    : SSPBUF holds the byte read from the MCP device
' Notes     : Uses  the second parameter to return the value
'
$define MCP23S17_Read(pRegAddress, pReturn) '
    WREG = pRegAddress '
    _Read_MCP          '
    pReturn = WREG

_Read_MCP Macro-
    GoSub __Read_MCP
Endm

#ifMacro- _Read_MCP, _ReadPortA, _ReadPortB
__Read_MCP:
    bRegAddress = WREG
    Clear MCP_CS
    DelayCS 1
    mSPI_Write(cControlByteMaskRead)
    mSPI_Write(bRegAddress)
    WREG = mSPI_Read()
    Set MCP_CS
    Return
#endIfMacro-

'----------------------------------------------------------------------------------------------------
' Reads PORTA
' Input     : None
' Output    : WREG holds the contents of PORTA
' Notes     : Saves the port value to a shadow variable
'
$define MCP23S17_ReadPortA() _ReadPortA

_ReadPortA Macro- \Byte
    WREG = cGPIOA
    GoSub __Read_MCP
    bShadowDataRegA = WREG
    Return_Wreg
Endm
'----------------------------------------------------------------------------------------------------
' Reads PORTB
' Input     : None
' Output    : WREG holds the contents of PORTB
' Notes     : Saves the port value to a shadow variable
'
$define MCP23S17_ReadPortB() _ReadPortB

_ReadPortB Macro- \Byte
    WREG = cGPIOB
    GoSub __Read_MCP
    bShadowDataRegB = WREG
    Return_Wreg
Endm

'----------------------------------------------------------------------------------------------------
' Writes PORTA Latch register
' Input     : pDataByte holds the byte to place on the Latch
' Output    : None
' Notes     : Saves the port value to a shadow variable
'
$define MCP23S17_WritePortA(pDataByte) '
    bShadowDataRegA = pDataByte        '
    MCP23S17_Write(cOLATA, pDataByte)

'----------------------------------------------------------------------------------------------------
' Writes PORTB Latch register
' Input     : pDataByte holds the Byte To place on the Latch
' Output    : None
' Notes     : Saves the port value to a shadow variable
'
$define MCP23S17_WritePortB(pDataByte) '
    bShadowDataRegB = pDataByte        '
    MCP23S17_Write(cOLATB, pDataByte)

'----------------------------------------------------------------------------------------------------
' Sets PORTA pin's direction
' Input     : pDirection holds the direction(0=Output 1=Input )
' Output    : None
' Notes     : Saves the TRIS value to a shadow variable
'
$define MCP23S17_SetTrisPortA(pDirection) '
    MCP23S17_Write(cIODIRA, pDirection)   '
    bShadowDirRegA = pDirection

'----------------------------------------------------------------------------------------------------
' Sets PORTB pin's direction
' Input     : pDirection holds the direction(0=Output 1=Input )
' Output    : None
' Notes     : Saves the TRIS value to a shadow variable
'
$define MCP23S17_SetTrisPortB(pDirection) '
    MCP23S17_Write(cIODIRB, pDirection)   '
    bShadowDirRegB = pDirection

'----------------------------------------------------------------------------------------------------
' Write to a bit of PORTA
' Input     : pPin Output pin to set or clear
'             pHiLo 1 = high, 0 = low
' Output    : None
' Notes     : Uses the shadow port to read the previous state
'
$define MCP23S17_WritePinPortA(pPin,pHiLo) '
    bMCPtemp = pHiLo '
    WREG = pPin      '
    _WritePinPortA

_WritePinPortA Macro-
    GoSub __WritePinPortA
Endm

#ifMacro- _WritePinPortA
__WritePinPortA:
    PRODL = WREG
    PRODH =  1 << PRODL
    If bMCPtemp = 1 Then
        PRODH = PRODH | bShadowDataRegA
    Else
        PRODH = ~PRODH
        PRODH = PRODH & bShadowDataRegA
    EndIf
    MCP23S17_Write(cOLATA, PRODH)
    Return
#endIfMacro

'----------------------------------------------------------------------------------------------------
' Write to a bit of PORTB
' Input     : pPin Output pin to set or clear
'             pHiLo 1 = high, 0 = low
' Output    : None
' Notes     : Uses the shadow port to read the previous state
'
$define MCP23S17_WritePinPortB(pPin,pHiLo) '
    bMCPtemp = pHiLo    '
    WREG = pPin         '
    _WritePinPortB

_WritePinPortB Macro-
    GoSub __WritePinPortB
Endm

#ifMacro- _WritePinPortB
__WritePinPortB:
    PRODL = WREG
    PRODH =  1 << PRODL
    If bMCPtemp = 1 Then
        PRODH = PRODH | bShadowDataRegB
    Else
        PRODH = ~PRODH
        PRODH = PRODH & bShadowDataRegB
    EndIf
    MCP23S17_Write(cOLATB, PRODH)
    Return
#endIfMacro-

'----------------------------------------------------------------------------------------------------
' Sets or clears the TRIS of a Single Pin from PORTA pins
' Parameters : pPin TRIS pin to set or clear (0 - 7)
'              pDirection 1 = Input, 0 = Output
' Input     : None
' Output    : None
' Notes     : None
'
$define MCP23S17_WritePinTrisA(pPin,pDirection) '
    bMCPtemp = pDirection   '
    WREG = pPin             '
    _WritePinTrisA

_WritePinTrisA Macro-
    GoSub __WritePinTrisA
Endm

#ifMacro- _WritePinTrisA
__WritePinTrisA:
    PRODL = WREG
    PRODH =  1 << PRODL
    If bMCPtemp = 1 Then
        PRODH = PRODH | bShadowDirRegA
    Else
        PRODH = ~PRODH
        PRODH = PRODH & bShadowDirRegA
    EndIf
    bShadowDirRegA = PRODH
    MCP23S17_Write(cIODIRA, PRODH)
    Return
#endIfMacro
'----------------------------------------------------------------------------------------------------
' Sets or clears the TRIS of a Single Pin from PORTB pins
' Input     : pPin holds the TRIS pin to set or clear (0 - 7)
'             pDirection 1 = Input, 0 = Output
' Output    : None
' Notes     : None
'
$define MCP23S17_WritePinTrisB(pPin,pDirection) '
    bMCPtemp = pDirection   '
    WREG = pPin             '
    _WritePinTrisB

_WritePinTrisB Macro-
    GoSub __WritePinTrisB
Endm

#ifdef _WritePinTrisB#Req
__WritePinTrisB:
    PRODL = WREG
    PRODH =  1 << PRODL
    If bMCPtemp = 1 Then
        PRODH = PRODH | bShadowDirRegB
    Else
        PRODH = ~PRODH
        PRODH = PRODH & bShadowDirRegB
    EndIf
    bShadowDirRegB = PRODH
    MCP23S17_Write(cIODIRB, PRODH)
    Return
#endif

'----------------------------------------------------------------------------------------------------
' Gets PORTA pins direction 0=Output 1=Input
' Input     : None
' Output    :
' Notes     : Reads the shadow register instead of reading the register
'
$define MCP23S17_ReadTrisPortA() bShadowDirRegA

'----------------------------------------------------------------------------------------------------
' Gets PORTB pins direction 0=Output 1=Input
' Input     : None
' Output    :
' Notes     : Reads the shadow register instead of reading the register
'
$define MCP23S17_ReadTrisPortB() bShadowDirRegB

'----------------------------------------------------------------------------------------------------
' Set the Pullups resistors  for any Pin of PORTA
' Input     : None
' Output    :
' Notes     : Pullups can be enabled/disabled independently for any pin of the port 0=Disabled, 1=Enabled
'
$define MCP23S17_SetPullUpsPortA(pPullUps) MCP23S17_Write(cGPPUA, pPullUps)

'----------------------------------------------------------------------------------------------------
' Sets the Pullups resistors  for any Pin of PORTB
' Input     : None
' Output    :
' Notes     : Pullups can be enabled/disabled independently for any pin of the port 0=Disabled, 1=Enabled
'
$define MCP23S17_SetPullUpsPortB(pPullUps) MCP23S17_Write(cGPPUB, pPullUps)

'----------------------------------------------------------------------------------------------------
' Initialise the MCP23S17
' Chip_Address must be forced to "000"
' regardless of actual Mcpchip_Address value.
' After initialisation, once IOCON.HAEN is enabled,
' Mcpchip_Address is taken into account by the chip.
'
$define MCP23S17_Init() _MCP23S17_Init

_MCP23S17_Init Macro-
    GoSub __MCP23S17_Init
    Endm

#ifMacro- _MCP23S17_Init
__MCP23S17_Init:
    High MCP_CS
$ifdef MCP_RST                                   ' Is a reset line being used?
    Low MCP_RST                                  ' Yes. So pull the reset line low
    DelayMS 10                                   ' Wait 10 milliseconds
    Set MCP_RST                                  ' Bring the device out of reset
$endif

    mSPI_Open(cSPI_FOSC_16, cSPI_MODE_00, cSPI_SMPEND) ' Open the SPI peripheral
    Clear MCP_CS
    DelayCS 1
    mSPI_Write(cControlByteMaskWrite & %11110001)  ' Force ChipAdd to 000
    mSPI_Write(cIOCONA)                            ' Writing to cIOCONA writes also to cIOCONB
    mSPI_Write(cControl_IOCON)
    Set MCP_CS
    bShadowDirRegA = %11111111
    bShadowDirRegB = %11111111
    bShadowDataRegA = %00000000
    bShadowDataRegB = %00000000
    DelayMS 10
    Return
#endIfMacro-

'----------------------------------------------------------------------------------------------------
MCP23S17_Main:
    MCP23S17_Init()

$endif



